# Copyright (C) 2009 Eugene Tarasenko, Alexander Belchenko
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Support external branches like svn:externals.

Using hooks, this plugin supports external branches for these commands:

* branch
* checkout
* pull
* push
* update.

Works from any client including bzr, QBzr, TortoiseBzr and Bazaar Explorer.

To define an external branch, add a line to the ``.bzrmeta/externals`` file
using the following format::

  url directory [revisionspec]

Example definitions are::

  bzr://example.com/repos/common/foo common/foo revno:100
  # relative to scheme
  //example.com/repos/common/foo common/foo
  # relative to the server hostname
  /repos/common/foo common/foo
  # relative to parent directory
  ../../common/foo common/foo

This plugin also provides a command to execute an arbitrary bzr or shell
command for each branch in the local workspace, see also
``bzr help externals-command``.

Known issues:

* Does not support ``^/`` expression in relative urls.
* pull command does not overwrite your newer working tree if you use a custom
  revision.
* update command does not support a custom revision and does not check
  externals if the branch tip is not changed.
* Does not support lightweight checkout.
"""

from bzrlib import branch, progress
from bzrlib.commands import plugin_cmds

plugin_name = 'externals'
version_info = (1, 2, 0, 'final', 0)

def install_hooks():
    def lazy_post_change_branch_tip_hook(params):
        from bzrlib.plugins.externals import externals
        return externals.post_change_branch_tip_hook(params)

    branch.Branch.hooks.install_named_hook(
        'post_change_branch_tip',
        lazy_post_change_branch_tip_hook,
        'branch/checkout external branches')

    def lazy_post_pull_hook(result):
        from bzrlib.plugins.externals import externals
        return externals.post_pull_hook(result)

    branch.Branch.hooks.install_named_hook(
        'post_pull',
        lazy_post_pull_hook,
        'pull/update external branches')

    def lazy_post_push_hook(result):
        from bzrlib.plugins.externals import externals
        return externals.post_push_hook(result)

    branch.Branch.hooks.install_named_hook(
        'post_push',
        lazy_post_push_hook,
        'push external branches')


if hasattr(branch.Branch.hooks, "install_named_hook"):
    install_hooks()

plugin_cmds.register_lazy('cmd_externals_command', ['ecmd'],
    'bzrlib.plugins.externals.commands')

# monkey patching for sets branch.parent before call post_change_branch_tip

def new_copy_content_into(self, destination, revision_id=None):
    destination.set_parent(self.base)
    old_copy_content_into(self, destination, revision_id)

old_copy_content_into = branch.Branch.copy_content_into
branch.Branch.copy_content_into = new_copy_content_into

# monkey patching for fix trash after pull command

def new_finished(self):
    old_finished(self)
    # clear line
    from bzrlib import ui
    ui.ui_factory.clear_term()

old_finished = progress.ProgressTask.finished
progress.ProgressTask.finished = new_finished
