# Copyright (C) 2010 Eugene Tarasenko, Marius Kruger
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os, os.path

from bzrlib import bzrdir
from bzrlib.osutils import getcwd, relpath
from bzrlib.commands import Command, run_bzr_catch_user_errors
from bzrlib.option import Option
from bzrlib.trace import is_quiet, note
from bzrlib.plugins.externals import externals

_main_cwd = getcwd()
_main_base = None

class cmd_externals_command(Command):
    """Run some command(s) for this and external branches.

    This command run your COMMAND at first for each external branches,
    then for main branch if not use ``--externals-only`` option.  
    For using an additional command option please use ``--`` after main command.

    :Examples:
      bzr ecmd status
      bzr ecmd -- log -r-1
    """
    aliases = ['ecmd']
    takes_args = ['command+']
    takes_options = [
        Option('externals-only', short_name='e',
            help='Don\'t run on the main branch.'),
        Option('shell', short_name='s',
            help='Run shell commands instead of bzr commands.'),
        'dry-run']

    def run(self, command_list, externals_only=False, shell=False, dry_run=False):
        # TODO: support setting the base location with -d
        externals.disable_hooks = True
        main_bzrdir, relpath = bzrdir.BzrDir.open_containing('.')
        branch = main_bzrdir.open_branch()

        global _main_base
        if _main_base is None:
            _main_base = branch.base
            
        ex = externals.Externals(branch, branch.last_revision())
        if ex.read_config():
            ecmd = ['ecmd']
            ex.adjust_verbosity(ecmd)
            for location, dir, revision, bound in ex.branch_iterator():
                self._run_bzr_or_shell(ex.cwd, dir,
                    ecmd + ['--'] + command_list, shell, dry_run)
        if not externals_only:
            if not is_quiet():
                if branch.base != _main_base:
                    note('Run command in external: ' + self._relpath(ex.root))
                else:
                    note('Run command in main branch:')
            ex.adjust_verbosity(command_list)
            self._run_bzr_or_shell(ex.root, '', command_list, shell, dry_run)
        os.chdir(ex.cwd)
                    
    def _run_bzr_or_shell(self, path, dir, command_list, shell, dry_run):
        if not dry_run:
            os.chdir(os.path.join(path, dir))
            if shell:
                os.system(' '.join(command_list))
            else:
                run_bzr_catch_user_errors(command_list)

    def _relpath(self, path):
        try:
            global _main_cwd
            return relpath(_main_cwd, path)
        except:
            pass
        return path
