# Copyright (C) 2009 Eugene Tarasenko, Alexander Belchenko
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os

from bzrlib.urlutils import join, local_path_from_url
from bzrlib.osutils import check_legal_path, getcwd, get_user_encoding, isdir, isfile, pathjoin, relpath
from bzrlib.commands import run_bzr, shlex_split_unicode
from bzrlib.trace import is_quiet, is_verbose, note

CONFIG_PATH = '.bzrmeta/externals'
disable_hooks = False # for run ecmd command

class Externals:
    
    def __init__(self, branch, revid):
        self.branch = branch
        self.revid = revid
        self.root = local_path_from_url(branch.base)
        self.cwd = getcwd()
        self.config = []

    def _set_config(self, text):
        lines = text.splitlines()
        for line in lines:
            if not line.startswith('#'):
                # function shlex.split not support windows path separator '\'
                line = line.decode('utf-8').replace('\\', '/')
                if line not in self.config:
                    self.config.append(line)

    def _read_config_from_repository(self):
        rev_tree = self.branch.repository.revision_tree(self.revid)
        file_id = rev_tree.path2id(CONFIG_PATH)
        if not file_id:
            # there is no .bzrmeta/externals file in repository
            return False
        rev_tree.lock_read()
        try:
            text = rev_tree.get_file_text(file_id)
            self._set_config(text)
        finally:
            rev_tree.unlock()
        return True

    def _read_config_from_directory(self):
        path = pathjoin(self.root, CONFIG_PATH)
        if not isfile(path):
            # file not exist in directory
            return False
        f = open(path, 'rU')
        try:
            self._set_config(f.read())
        finally:
            f.close()
        return True

    def read_config(self):
        in_repo = self._read_config_from_repository()
        in_dir = self._read_config_from_directory()
        return in_repo or in_dir

    def _relpath(self, path):
        try:
            return relpath(self.cwd, path)
        except:
            pass
        return path

    @staticmethod
    def _urljoin(base, relative):
        if relative.startswith('//'):
            # urlutils.join not supports relative urls start with '//'
            scheme = base.partition('://')
            return scheme[0] + ':' + relative
        else:
            return join(base, relative)

    @staticmethod
    def _report(cmd):
        if not is_quiet():
            note('External ' + ' '.join(cmd))

    @staticmethod
    def adjust_verbosity(cmd):
        if is_quiet():
            cmd += ['-q']
        if is_verbose():
            cmd += ['-v']

    def branch_iterator(self, target_root=None):
        if self.config is None:
            # branch not have externals configuration
            return

        bound = True
        if not target_root:
            target_root = self.branch.get_bound_location()
            if not target_root:
                bound = False
                target_root = self.branch.get_parent()
                if not target_root:
                    # support new braches with no parent yet
                    target_root = self.branch.base

        for line in self.config:
            arg = shlex_split_unicode(line) # url directory [revision]
            location = self._urljoin(target_root, arg[0])
            if target_root.startswith('file:///'):
                # try to pull externals from the parent for the feature branch
                path = pathjoin(local_path_from_url(target_root), arg[1])
                if isdir(path):
                    location = self._relpath(path)
                else:
                    # parent is local master branch
                    if location.startswith('file:///'):
                        location = self._relpath(local_path_from_url(location))

            check_legal_path(arg[1])
            rel_path = self._relpath(pathjoin(self.root, arg[1]))
            
            revision = None
            if len(arg) > 2:
                revision = arg[2]
            yield location, rel_path, revision, bound

    def pull(self):
        for location, path, revision, bound in self.branch_iterator():
            if location == path:
                # not create feature branch for directory above the root
                continue

            # select what do it
            if isdir(pathjoin(path, '.bzr')) or isdir(pathjoin(path, '.svn')):
                if bound:
                    cmd = ['update', path]
                else:
                    cmd = ['pull', location, '--directory', path]
            else:
                if bound:
                    cmd = ['checkout', location, path]
                else:
                    cmd = ['branch', location, path]

                # command branch don't create recursive directory
                dirs = path.rpartition('/')
                if dirs[0] != '' and not isdir(dirs[0]):
                    os.makedirs(dirs[0].encode(get_user_encoding()))

            # if use revision options but not for 'update'
            if revision is not None and cmd[0] != 'update':
                cmd += ['--revision', revision]

            self.adjust_verbosity(cmd)
            self._report(cmd)
            run_bzr(cmd)

    def push(self, target):
        for location, path, revision, bound in self.branch_iterator(target):
            if location == path:
                # don't push into itself
                continue

            # XXX: maybe he should rather decorate the push command
            # so that we can get the commandline args
            # alternatively the plugin infrastructure must provide it to us?!
            cmd = ['push', location, '--directory', path, '--no-strict']

            if revision is not None:
                # not push if use revision 
                continue

            self.adjust_verbosity(cmd)
            self._report(cmd)
            run_bzr(cmd)

    def add_to_ignore(self):
        from bzrlib import IGNORE_FILENAME
        f = open(pathjoin(self.root, IGNORE_FILENAME), 'a+')
        try:
            patterns = set()
            line = '\n'
            for line in f:
                if not line.startswith('#'):
                    patterns.add(line.decode('utf-8').strip())

            # add at end of file the char '\n' if needed
            if not line.endswith('\n'):
                f.write('\n')

            # not used WorkingTree.is_ignored because tree is locked
            from bzrlib.globbing import Globster
            ignored = Globster(patterns)

            for line in self.config:
                arg = shlex_split_unicode(line)
                if not ignored.match(arg[1]):
                    f.write('./%s\n' % arg[1].encode('utf-8'))
        finally:
            f.close()


def post_change_branch_tip_hook(params):
    if not disable_hooks:
        if params.old_revno == 0 and params.branch.base.startswith('file:///'):
            # only branch or checkout commands
            ex = Externals(params.branch, params.new_revid)
            if ex.read_config():
                ex.pull()

def post_pull_hook(result):
    if not disable_hooks:
        if result.old_revno != 0:
            # only pull or update commands
            ex = Externals(result.target_branch, result.new_revid)
            if ex.read_config():
                ex.pull()
                ex.add_to_ignore()

def post_push_hook(result):
    if not disable_hooks:
        ex = Externals(result.source_branch, result.new_revid)
        if ex.read_config():
            ex.push(result.target_branch.base)
